@@ -1,9 +1,9 @@
 /*
+    SPDX-FileCopyrightText: 2016 The Qt Company Ltd. <https://www.qt.io/licensing/>
     SPDX-FileCopyrightText: 2017 Marco Martin <mart@kde.org>
     SPDX-FileCopyrightText: 2017 David Edmundson <davidedmundson@kde.org>
-    SPDX-FileCopyrightText: 2016 The Qt Company Ltd. <https://www.qt.io/licensing/>
-
-    This file is part of the Qt Quick Controls module of the Qt Toolkit.
+    SPDX-FileCopyrightText: 2019 David Redondo <david@david-redondo.de>
+    SPDX-FileCopyrightText: 2023 ivan tkachenko <me@ratijas.tk>
 
     SPDX-License-Identifier: LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only OR LicenseRef-KFQF-Accepted-GPL OR LicenseRef-Qt-Commercial
 */
@@ -26,15 +26,15 @@
 #include <Kirigami/PlatformTheme>
 #include <Kirigami/TabletModeWatcher>
 
-QStyle *KQuickStyleItem::s_style = nullptr;
+std::unique_ptr<QStyle> KQuickStyleItem::s_style = nullptr;
 
 QStyle *KQuickStyleItem::style()
 {
-    if (!qobject_cast<QApplication *>(QCoreApplication::instance())) {
-        return s_style;
+    if (s_style) {
+        return s_style.get();
+    } else {
+        return qApp->style();
     }
-    QStyle *style = qApp->style();
-    return style ? style : s_style;
 }
 
 KQuickStyleItem::KQuickStyleItem(QQuickItem *parent)
@@ -91,16 +91,12 @@ KQuickStyleItem::KQuickStyleItem(QQuickItem *parent)
         if (style) {
             connect(style, &QObject::destroyed, this, &KQuickStyleItem::styleChanged);
         }
-    } else {
+    } else if (!s_style) {
         // We are not a QApplication.  Create an internal copy of the configured
         // desktop style, to be used for metrics, options and painting.
         KSharedConfig::Ptr kdeglobals = KSharedConfig::openConfig();
         KConfigGroup cg(kdeglobals, "KDE");
-        auto style = s_style;
-        s_style = QStyleFactory::create(cg.readEntry("widgetStyle", QStringLiteral("Fusion")));
-        if (style) {
-            delete style;
-        }
+        s_style.reset(QStyleFactory::create(cg.readEntry("widgetStyle", QStringLiteral("Fusion"))));
     }
 
     m_font = qApp->font();
@@ -173,6 +169,7 @@ void KQuickStyleItem::initStyleOption()
     QString sizeHint = m_hints.value(QStringLiteral("size")).toString();
 
     bool needsResolvePalette = true;
+    bool preventMirroring = false;
 
     switch (m_itemType) {
     case Button: {
@@ -296,6 +293,7 @@ void KQuickStyleItem::initStyleOption()
         opt->fontMetrics = QFontMetrics(font);
         break;
     }
+    case DelayButton:
     case ToolButton: {
         if (!m_styleoption) {
             m_styleoption = new QStyleOptionToolButton();
@@ -311,7 +309,10 @@ void KQuickStyleItem::initStyleOption()
         opt->activeSubControls = QStyle::SC_ToolButton;
         opt->text = text();
 
-        opt->icon = iconFromIconProperty();
+        if (m_iconDirty || opt->icon.isNull()) {
+            opt->icon = iconFromIconProperty();
+            m_iconDirty = false;
+        }
 
         auto iconSize = QSize(m_properties[QStringLiteral("iconWidth")].toInt(), m_properties[QStringLiteral("iconHeight")].toInt());
         if (iconSize.isEmpty()) {
@@ -363,6 +364,23 @@ void KQuickStyleItem::initStyleOption()
         QStyleOptionTab *opt = qstyleoption_cast<QStyleOptionTab *>(m_styleoption);
         opt->text = text();
 
+        if (m_iconDirty || opt->icon.isNull()) {
+            opt->icon = iconFromIconProperty();
+            m_iconDirty = false;
+        }
+
+        auto iconSize = QSize(m_properties[QStringLiteral("iconWidth")].toInt(), m_properties[QStringLiteral("iconHeight")].toInt());
+        if (iconSize.isEmpty()) {
+            int e = KQuickStyleItem::style()->pixelMetric(QStyle::PM_TabBarIconSize, m_styleoption, nullptr);
+            if (iconSize.width() <= 0) {
+                iconSize.setWidth(e);
+            }
+            if (iconSize.height() <= 0) {
+                iconSize.setHeight(e);
+            }
+        }
+        opt->iconSize = iconSize;
+
         if (m_properties.value(QStringLiteral("hasFrame")).toBool()) {
             opt->features |= QStyleOptionTab::HasFrame;
         }
@@ -401,7 +419,7 @@ void KQuickStyleItem::initStyleOption()
         QStyleOptionFrame *opt = qstyleoption_cast<QStyleOptionFrame *>(m_styleoption);
         opt->frameShape = QFrame::StyledPanel;
         opt->lineWidth = KQuickStyleItem::style()->pixelMetric(QStyle::PM_DefaultFrameWidth, m_styleoption, nullptr);
-        opt->midLineWidth = KQuickStyleItem::style()->pixelMetric(QStyle::PM_DefaultFrameWidth, m_styleoption, nullptr);
+        opt->midLineWidth = opt->lineWidth;
         break;
     }
     case FocusRect: {
@@ -499,7 +517,12 @@ void KQuickStyleItem::initStyleOption()
                     opt->checkType = exclusive.toBool() ? QStyleOptionMenuItem::Exclusive : QStyleOptionMenuItem::NonExclusive;
                 }
             }
-            opt->icon = iconFromIconProperty();
+
+            if (m_iconDirty || opt->icon.isNull()) {
+                opt->icon = iconFromIconProperty();
+                m_iconDirty = false;
+            }
+
             setProperty("_q_showUnderlined", m_hints[QStringLiteral("showUnderlined")].toBool());
 
             const QFont font = qApp->font(m_itemType == ComboBoxItem ? "QComboMenuItem" : "QMenu");
@@ -524,7 +547,10 @@ void KQuickStyleItem::initStyleOption()
         }
         opt->text = text();
 
-        opt->icon = iconFromIconProperty();
+        if (m_iconDirty || opt->icon.isNull()) {
+            opt->icon = iconFromIconProperty();
+            m_iconDirty = false;
+        }
 
         auto iconSize = QSize(m_properties[QStringLiteral("iconWidth")].toInt(), m_properties[QStringLiteral("iconHeight")].toInt());
         if (iconSize.isEmpty()) {
@@ -586,11 +612,12 @@ void KQuickStyleItem::initStyleOption()
 
         QStyleOptionSpinBox *opt = qstyleoption_cast<QStyleOptionSpinBox *>(m_styleoption);
         opt->frame = true;
-        opt->subControls = QStyle::SC_SpinBoxFrame | QStyle::SC_SpinBoxEditField;
         if (value() & 0x1) {
             opt->activeSubControls = QStyle::SC_SpinBoxUp;
         } else if (value() & (1 << 1)) {
             opt->activeSubControls = QStyle::SC_SpinBoxDown;
+        } else {
+            opt->activeSubControls = QStyle::SC_None;
         }
         opt->subControls = QStyle::SC_All;
         opt->stepEnabled = {};
@@ -691,6 +718,10 @@ void KQuickStyleItem::initStyleOption()
         opt->maximum = qMax(0, maximum());
         opt->pageStep = qMax(0, int(horizontal() ? width() : height()));
         opt->orientation = horizontal() ? Qt::Horizontal : Qt::Vertical;
+        if (horizontal()) {
+            // mirrored horizontal scrollbars are not something you wanna interact with
+            preventMirroring = true;
+        }
         opt->sliderPosition = value();
         opt->sliderValue = value();
         opt->activeSubControls = (activeControl() == QLatin1String("up")) ? QStyle::SC_ScrollBarSubLine
@@ -702,7 +733,6 @@ void KQuickStyleItem::initStyleOption()
             opt->state |= QStyle::State_On;
         }
 
-        opt->sliderValue = value();
         opt->subControls = QStyle::SC_All;
 
         setTransient(KQuickStyleItem::style()->styleHint(QStyle::SH_ScrollBar_Transient, m_styleoption));
@@ -722,7 +752,7 @@ void KQuickStyleItem::initStyleOption()
 
     m_styleoption->styleObject = this;
     const auto mirror = m_control == nullptr ? qApp->layoutDirection() == Qt::RightToLeft : m_control->property("mirrored").toBool();
-    m_styleoption->direction = mirror ? Qt::RightToLeft : Qt::LeftToRight;
+    m_styleoption->direction = (mirror && !preventMirroring) ? Qt::RightToLeft : Qt::LeftToRight;
 
     int w = m_textureWidth > 0 ? m_textureWidth : width();
     int h = m_textureHeight > 0 ? m_textureHeight : height();
@@ -824,6 +854,7 @@ const char *KQuickStyleItem::classNameForItem() const
         return "QComboMenuItem";
     case ToolBar:
         return "";
+    case DelayButton:
     case ToolButton:
         return "QToolButton";
     case Tab:
@@ -1015,21 +1046,21 @@ QSize KQuickStyleItem::sizeFromContents(int width, int height)
     QSize size;
     switch (m_itemType) {
     case RadioButton:
-        size = KQuickStyleItem::style()->sizeFromContents(QStyle::CT_RadioButton, m_styleoption, QSize(width, height));
-        break;
     case CheckBox: {
+        auto contentType = (m_itemType == RadioButton) ? QStyle::CT_RadioButton : QStyle::CT_CheckBox;
         QStyleOptionButton *btn = qstyleoption_cast<QStyleOptionButton *>(m_styleoption);
         QSize contentSize = btn->fontMetrics.size(Qt::TextShowMnemonic, btn->text);
         if (!btn->icon.isNull()) {
             contentSize.setWidth(contentSize.width() + btn->iconSize.width());
             contentSize.setHeight(std::max(contentSize.height(), btn->iconSize.height()));
         }
-        size = KQuickStyleItem::style()->sizeFromContents(QStyle::CT_CheckBox, m_styleoption, contentSize);
+        size = KQuickStyleItem::style()->sizeFromContents(contentType, m_styleoption, contentSize);
         break;
     }
     case ToolBar:
         size = QSize(200, styleName().contains(QLatin1String("windows")) ? 30 : 42);
         break;
+    case DelayButton:
     case ToolButton: {
         QStyleOptionToolButton *btn = qstyleoption_cast<QStyleOptionToolButton *>(m_styleoption);
         int w = 0;
@@ -1470,6 +1501,8 @@ void KQuickStyleItem::setElementType(const QString &str)
         m_itemType = MenuBar;
     } else if (str == QLatin1String("menubaritem")) {
         m_itemType = MenuBarItem;
+    } else if (str == QLatin1String("delaybutton")) {
+        m_itemType = DelayButton;
     } else {
         m_itemType = Undefined;
     }
@@ -1514,8 +1547,7 @@ QRectF KQuickStyleItem::subControlRect(const QString &subcontrolString)
         QStyle::ComplexControl control = QStyle::CC_ScrollBar;
         if (subcontrolString == QLatin1String("slider")) {
             subcontrol = QStyle::SC_ScrollBarSlider;
-        }
-        if (subcontrolString == QLatin1String("groove")) {
+        } else if (subcontrolString == QLatin1String("groove")) {
             subcontrol = QStyle::SC_ScrollBarGroove;
         } else if (subcontrolString == QLatin1String("handle")) {
             subcontrol = QStyle::SC_ScrollBarSlider;
@@ -1575,6 +1607,8 @@ void KQuickStyleItem::paint(QPainter *painter)
             font = qApp->font("QMiniFont");
         } else if (m_styleoption->state & QStyle::State_Small) {
             font = qApp->font("QSmallFont");
+        } else {
+            font = QApplication::font(classNameForItem());
         }
         painter->setFont(font);
     }
@@ -1752,6 +1786,30 @@ void KQuickStyleItem::paint(QPainter *painter)
         }
         break;
     }
+    case DelayButton: { // Adapted from Spectacle's ProgressButton made by David Redondo.
+        // Draw Button without text and icon, note the missing text and icon in options
+        QStyleOption baseStyleOptions = *m_styleoption;
+        baseStyleOptions.state.setFlag(QStyle::State_Enabled, !baseStyleOptions.state.testFlag(QStyle::State_Sunken));
+        KQuickStyleItem::style()->drawPrimitive(QStyle::PE_PanelButtonTool, &baseStyleOptions, painter);
+        qreal progress = qreal(value()) / qreal(maximum());
+        if (!qFuzzyIsNull(progress)) {
+            // Draw overlay
+            QStyleOption overlayOption;
+            overlayOption.palette = m_styleoption->palette;
+            overlayOption.palette.setBrush(QPalette::Button, m_styleoption->palette.highlight());
+            overlayOption.rect = m_styleoption->direction == Qt::LeftToRight ?
+                QRect(0, 0, m_styleoption->rect.width() * progress, m_styleoption->rect.height())
+                : QRect(QPoint(m_styleoption->rect.width() * (1 - progress), 0), m_styleoption->rect.size());
+            overlayOption.state.setFlag(QStyle::State_Sunken, m_styleoption->state.testFlag(QStyle::State_Sunken));
+            painter->setCompositionMode(QPainter::CompositionMode_SourceAtop);
+            painter->setOpacity(0.5);
+            KQuickStyleItem::style()->drawPrimitive(QStyle::PE_PanelButtonTool, &overlayOption, painter);
+        }
+        // Finally draw text and icon and outline
+        painter->setOpacity(1);
+        KQuickStyleItem::style()->drawControl(QStyle::CE_ToolButtonLabel, m_styleoption, painter);
+        break;
+    }
     default:
         break;
     }
@@ -1872,12 +1930,13 @@ QSGNode *KQuickStyleItem::updatePaintNode(QSGNode *node, UpdatePaintNodeData *)
     }
 
 #ifdef QSG_RUNTIME_DESCRIPTION
-    qsgnode_set_description(styleNode, QString::fromLatin1("%1:%2, '%3'").arg(styleName()).arg(elementType()).arg(text()));
+    qsgnode_set_description(styleNode, QString::fromLatin1("%1:%2, '%3'").arg(styleName(), elementType(), text()));
 #endif
 
     styleNode->setTexture(window()->createTextureFromImage(m_image, QQuickWindow::TextureCanUseAtlas));
     styleNode->setBounds(boundingRect());
-    styleNode->setDevicePixelRatio(window()->devicePixelRatio());
+    styleNode->setDevicePixelRatio(window()->effectiveDevicePixelRatio());
+
     styleNode->setPadding(m_border.left(), m_border.top(), m_border.right(), m_border.bottom());
     styleNode->update();
 
@@ -1887,7 +1946,7 @@ QSGNode *KQuickStyleItem::updatePaintNode(QSGNode *node, UpdatePaintNodeData *)
 void KQuickStyleItem::updatePolish()
 {
     if (isVisible() && width() >= 1 && height() >= 1) { // Note these are reals so 1 pixel is minimum
-        const qreal devicePixelRatio = window() ? window()->devicePixelRatio() : qApp->devicePixelRatio();
+        const qreal devicePixelRatio = window() ? window()->effectiveDevicePixelRatio() : qApp->devicePixelRatio();
         const QSize size = QSize(m_textureWidth > 0 ? m_textureWidth : width(), m_textureHeight > 0 ? m_textureHeight : height()) * devicePixelRatio;
 
         if (m_image.size() != size) {
@@ -1936,7 +1995,7 @@ bool KQuickStyleItem::eventFilter(QObject *watched, QEvent *event)
 
 void KQuickStyleItem::itemChange(QQuickItem::ItemChange change, const QQuickItem::ItemChangeData &value)
 {
-    if (change == QQuickItem::ItemVisibleHasChanged || change == QQuickItem::ItemEnabledHasChanged) {
+    if (change == QQuickItem::ItemVisibleHasChanged || change == QQuickItem::ItemEnabledHasChanged || change == QQuickItem::ItemDevicePixelRatioHasChanged) {
         polish();
     }
 
